from typing import Dict, List, Optional
from opentrons.util.async_helpers import ensure_yield

from opentrons.drivers.types import (
    AbsorbanceReaderLidStatus,
    AbsorbanceReaderDeviceState,
)

from .abstract import AbstractAbsorbanceReaderDriver


class SimulatingDriver(AbstractAbsorbanceReaderDriver):
    def __init__(self, serial_number: Optional[str] = None) -> None:
        self._serial_number = serial_number

    @ensure_yield
    async def get_device_info(self) -> Dict[str, str]:
        """Get device info"""
        return {
            "serial": self._serial_number if self._serial_number else "dummySerialAR",
            "model": "dummyModelAR",
            "version": "dummyVersionAR",
        }

    @ensure_yield
    async def connect(self) -> None:
        """Connect to absorbance reader"""
        pass

    @ensure_yield
    async def disconnect(self) -> None:
        """Disconnect from absorbance reader"""
        pass

    @ensure_yield
    async def is_connected(self) -> bool:
        """Check connection to absorbance reader"""
        return True

    @ensure_yield
    async def get_lid_status(self) -> AbsorbanceReaderLidStatus:
        return AbsorbanceReaderLidStatus.ON

    @ensure_yield
    async def get_available_wavelengths(self) -> List[int]:
        return [450, 570, 600, 650]

    @ensure_yield
    async def get_single_measurement(self, wavelength: int) -> List[float]:
        return [0.0]

    @ensure_yield
    async def initialize_measurement(self, wavelength: int) -> None:
        pass

    @ensure_yield
    async def get_status(self) -> AbsorbanceReaderDeviceState:
        return AbsorbanceReaderDeviceState.OK



import csv
from tempfile import NamedTemporaryFile
from typing import Optional, TextIO, Any, List

from opentrons.protocols.api_support.types import APIVersion

from .exceptions import ParameterValueError, RuntimeParameterRequired


# TODO(jbl 2024-08-02) This is a public facing class and as such should be moved to the protocol_api folder
class CSVParameter:
    def __init__(self, contents: Optional[bytes], api_version: APIVersion) -> None:
        self._contents = contents
        self._api_version = api_version
        self._file: Optional[TextIO] = None

    @property
    def file(self) -> TextIO:
        """Returns the file handler for the CSV file."""
        if self._file is None:
            text = self.contents
            temporary_file = NamedTemporaryFile("r+")
            temporary_file.write(text)
            temporary_file.flush()

            # Open a new file handler for the temporary file with read-only permissions and close the other
            self._file = open(temporary_file.name, "r")
            temporary_file.close()
        return self._file

    @property
    def file_opened(self) -> bool:
        """Return if a file handler has been opened for the CSV parameter."""
        return self._file is not None

    @property
    def contents(self) -> str:
        """Returns the full contents of the CSV file as a single string."""
        if self._contents is None:
            raise RuntimeParameterRequired(
                "CSV parameter needs to be set to a file for full analysis or run."
            )
        return self._contents.decode("utf-8")

    def parse_as_csv(
        self, detect_dialect: bool = True, **kwargs: Any
    ) -> List[List[str]]:
        """Returns a list of rows with each row represented as a list of column elements.

        If there is a header for the CSV that will be the first row in the list (i.e. `.rows()[0]`).
        All elements will be represented as strings, even if they are numeric in nature.
        """
        rows: List[List[str]] = []
        if detect_dialect:
            try:
                dialect = csv.Sniffer().sniff(self.contents[:1024])
                reader = csv.reader(self.contents.split("\n"), dialect, **kwargs)
            except (UnicodeDecodeError, csv.Error):
                raise ParameterValueError(
                    "Cannot parse dialect or contents from provided CSV contents."
                )
        else:
            try:
                reader = csv.reader(self.contents.split("\n"), **kwargs)
            except (UnicodeDecodeError, csv.Error):
                raise ParameterValueError("Cannot parse provided CSV contents.")
        try:
            for row in reader:
                rows.append(row)
        except (UnicodeDecodeError, csv.Error):
            raise ParameterValueError("Cannot parse provided CSV contents.")
        return rows



"""Commands that will cause inaccuracy or incorrect behavior but are still necessary."""

from .unsafe_blow_out_in_place import (
    UnsafeBlowOutInPlaceCommandType,
    UnsafeBlowOutInPlaceParams,
    UnsafeBlowOutInPlaceResult,
    UnsafeBlowOutInPlace,
    UnsafeBlowOutInPlaceCreate,
)
from .unsafe_drop_tip_in_place import (
    UnsafeDropTipInPlaceCommandType,
    UnsafeDropTipInPlaceParams,
    UnsafeDropTipInPlaceResult,
    UnsafeDropTipInPlace,
    UnsafeDropTipInPlaceCreate,
)

from .update_position_estimators import (
    UpdatePositionEstimatorsCommandType,
    UpdatePositionEstimatorsParams,
    UpdatePositionEstimatorsResult,
    UpdatePositionEstimators,
    UpdatePositionEstimatorsCreate,
)

__all__ = [
    # Unsafe blow-out-in-place command models
    "UnsafeBlowOutInPlaceCommandType",
    "UnsafeBlowOutInPlaceParams",
    "UnsafeBlowOutInPlaceResult",
    "UnsafeBlowOutInPlace",
    "UnsafeBlowOutInPlaceCreate",
    # Unsafe drop-tip command models
    "UnsafeDropTipInPlaceCommandType",
    "UnsafeDropTipInPlaceParams",
    "UnsafeDropTipInPlaceResult",
    "UnsafeDropTipInPlace",
    "UnsafeDropTipInPlaceCreate",
    # Update position estimate command models
    "UpdatePositionEstimatorsCommandType",
    "UpdatePositionEstimatorsParams",
    "UpdatePositionEstimatorsResult",
    "UpdatePositionEstimators",
    "UpdatePositionEstimatorsCreate",
]



import asyncio
from typing import Optional, Mapping, List, Dict, Any, Tuple

from opentrons.drivers.rpi_drivers.types import USBPort
from opentrons.drivers.absorbance_reader import (
    AbstractAbsorbanceReaderDriver,
    AbsorbanceReaderDriver,
    SimulatingDriver,
)
from opentrons.drivers.types import (
    AbsorbanceReaderLidStatus,
    AbsorbanceReaderPlatePresence,
    AbsorbanceReaderDeviceState,
)

from opentrons.hardware_control.execution_manager import ExecutionManager
from opentrons.hardware_control.poller import Reader
from opentrons.hardware_control.modules import mod_abc
from opentrons.hardware_control.modules.types import (
    ModuleType,
    AbsorbanceReaderStatus,
    LiveData,
    UploadFunction,
)


async def upload_func_placeholder(
    dfu_serial: str, firmware_file_path: str, kwargs: Dict[str, Any]
) -> Tuple[bool, str]:
    return False, "Not implemented"


class AbsorbanceReader(mod_abc.AbstractModule):
    """Hardware control interface for an attached Absorbance Reader module."""

    MODULE_TYPE = ModuleType.ABSORBANCE_READER

    @classmethod
    async def build(
        cls,
        port: str,
        usb_port: USBPort,
        execution_manager: ExecutionManager,
        hw_control_loop: asyncio.AbstractEventLoop,
        poll_interval_seconds: Optional[float] = None,
        simulating: bool = False,
        sim_model: Optional[str] = None,
        sim_serial_number: Optional[str] = None,
    ) -> "AbsorbanceReader":
        """Create an AbsorbanceReader."""
        driver: AbstractAbsorbanceReaderDriver
        if not simulating:
            driver = await AbsorbanceReaderDriver.create(
                port, usb_port, hw_control_loop
            )
        else:
            driver = SimulatingDriver(serial_number=sim_serial_number)
        module = cls(
            port=port,
            usb_port=usb_port,
            device_info=await driver.get_device_info(),
            execution_manager=execution_manager,
            driver=driver,
            hw_control_loop=hw_control_loop,
        )
        await module.setup()
        return module

    def __init__(
        self,
        port: str,
        usb_port: USBPort,
        driver: AbstractAbsorbanceReaderDriver,
        device_info: Mapping[str, str],
        execution_manager: ExecutionManager,
        hw_control_loop: asyncio.AbstractEventLoop,
    ) -> None:
        super().__init__(port, usb_port, execution_manager, hw_control_loop)
        self._driver = driver
        self._device_info = device_info

    async def deactivate(self, must_be_running: bool = True) -> None:
        """Deactivate the module.

        Contains an override to the `wait_for_is_running` step in cases where the
        module must be deactivated regardless of context."""
        await self._driver.disconnect()

    @property
    def status(self) -> AbsorbanceReaderStatus:
        """Return some string describing status."""
        return AbsorbanceReaderStatus.IDLE

    @property
    def lid_status(self) -> AbsorbanceReaderLidStatus:
        return AbsorbanceReaderLidStatus.UNKNOWN

    @property
    def plate_presence(self) -> AbsorbanceReaderPlatePresence:
        return AbsorbanceReaderPlatePresence.UNKNOWN

    @property
    def device_info(self) -> Mapping[str, str]:
        """Return a dict of the module's static information (serial, etc)"""
        return self._device_info

    @property
    def live_data(self) -> LiveData:
        """Return a dict of the module's dynamic information"""
        return {
            "status": self.status.value,
            "data": {
                "lidStatus": self.lid_status.value,
                "platePresence": self.plate_presence.value,
                "sampleWavelength": 400,
            },
        }

    @property
    def is_simulated(self) -> bool:
        """True if this is a simulated module."""
        return isinstance(self._driver, SimulatingDriver)

    @property
    def port(self) -> str:
        """The virtual port where the module is connected."""
        return self._port

    @property
    def usb_port(self) -> USBPort:
        """The physical port where the module is connected."""
        return self._usb_port

    async def wait_for_is_running(self) -> None:
        if not self.is_simulated:
            await self._execution_manager.wait_for_is_running()

    async def prep_for_update(self) -> str:
        """Prepare for an update.

        By the time this coroutine completes, the hardware should be ready
        to take an update. This implicitly tears down the module instance;
        it does not need to be either working or recoverable after this
        coroutine completes.

        :returns str: The port we're running on.
        """
        return ""

    def model(self) -> str:
        """A name for this specific module, matching module defs"""
        return "absorbanceReaderV1"

    @classmethod
    def name(cls) -> str:
        """A shortname used for matching usb ports, among other things"""
        return "absorbancereader"

    def firmware_prefix(self) -> str:
        """The prefix used for looking up firmware"""
        # TODO: (AA) This is a placeholder
        return ""

    def bootloader(self) -> UploadFunction:
        """Bootloader mode"""
        return upload_func_placeholder

    async def cleanup(self) -> None:
        """Clean up the module instance.

        Clean up, i.e. stop pollers, disconnect serial, etc in preparation for
        object destruction.
        """
        await self._driver.disconnect()

    async def set_sample_wavelength(self, wavelength: int) -> None:
        """Set the Absorbance Reader's active wavelength."""
        await self._driver.initialize_measurement(wavelength)

    async def start_measure(self, wavelength: int) -> List[float]:
        """Initiate a single measurement."""
        return await self._driver.get_single_measurement(wavelength)

    async def setup(self) -> None:
        """Setup the Absorbance Reader."""
        is_open = await self._driver.is_connected()
        if not is_open:
            await self._driver.connect()

    async def get_current_wavelength(self) -> None:
        """Get the Absorbance Reader's current active wavelength."""
        pass


class AbsorbanceReaderReader(Reader):
    device_state: AbsorbanceReaderDeviceState
    lid_status: AbsorbanceReaderLidStatus
    plate_presence: AbsorbanceReaderPlatePresence
    supported_wavelengths: List[int]

    def __init__(self, driver: AbsorbanceReaderDriver) -> None:
        self.device_state = AbsorbanceReaderDeviceState.UNKNOWN
        self.lid_status = AbsorbanceReaderLidStatus.UNKNOWN
        self.plate_presence = AbsorbanceReaderPlatePresence.UNKNOWN
        self.supported_wavelengths = []
        self._driver = driver

    async def read(self) -> None:
        await self.get_device_status()
        await self.get_supported_wavelengths()

    async def get_device_status(self) -> None:
        """Get the Absorbance Reader's current status."""
        self.device_state = await self._driver.get_status()

    async def get_supported_wavelengths(self) -> None:
        """Get the Absorbance Reader's supported wavelengths."""
        self.supported_wavelengths = await self._driver.get_available_wavelengths()

    async def get_lid_status(self) -> None:
        """Get the Absorbance Reader's lid status."""
        self.lid_status = await self._driver.get_lid_status()

    async def get_plate_presence(self) -> None:
        """Get the Absorbance Reader's plate presence."""
        self.plate_presence = await self._driver.get_plate_presence()



from __future__ import annotations

import asyncio
import re
from concurrent.futures.thread import ThreadPoolExecutor
from functools import partial
from typing import Optional, List, Dict


from .hid_protocol import (
    AbsorbanceHidInterface as AbsProtocol,
    ErrorCodeNames,
    DeviceStateNames,
)
from opentrons.drivers.types import (
    AbsorbanceReaderLidStatus,
    AbsorbanceReaderPlatePresence,
    AbsorbanceReaderDeviceState,
)
from opentrons.drivers.rpi_drivers.types import USBPort


SN_PARSER = re.compile(r'ATTRS{serial}=="(?P<serial>.+?)"')


class AsyncByonoy:
    """Async wrapper around Byonoy Device Library."""

    @staticmethod
    def match_device_with_sn(
        sn: str, devices: List[AbsProtocol.Device]
    ) -> AbsProtocol.Device:
        for device in devices:
            if device.sn == sn:
                return device
        raise RuntimeError(f"Unavailble module with serial number: {sn}")

    @staticmethod
    def serial_number_from_port(name: str) -> str:
        """
        Get the serial number from a port using pyusb.
        """
        import usb.core as usb_core  # type: ignore[import-untyped]

        port_numbers = tuple(int(s) for s in name.split("-")[1].split("."))
        device = usb_core.find(port_numbers=port_numbers)
        if device:
            return str(device.serial_number)
        raise RuntimeError(f"Could not find serial number for port: {name}")

    @classmethod
    async def create(
        cls,
        port: str,
        usb_port: USBPort,
        loop: Optional[asyncio.AbstractEventLoop] = None,
    ) -> AsyncByonoy:
        """
        Create an AsyncByonoy instance.

        Args:
            port: url or port name
            baud_rate: the baud rate
            timeout: optional timeout in seconds
            write_timeout: optional write timeout in seconds
            loop: optional event loop. if None get_running_loop will be used
            reset_buffer_before_write: reset the serial input buffer before
             writing to it
        """
        loop = loop or asyncio.get_running_loop()
        executor = ThreadPoolExecutor(max_workers=1)

        import pybyonoy_device_library as byonoy  # type: ignore[import-not-found]

        interface: AbsProtocol = byonoy

        device_sn = cls.serial_number_from_port(usb_port.name)
        found: List[AbsProtocol.Device] = await loop.run_in_executor(
            executor=executor, func=byonoy.byonoy_available_devices
        )
        device = cls.match_device_with_sn(device_sn, found)

        return cls(
            interface=interface,
            device=device,
            executor=executor,
            loop=loop,
        )

    def __init__(
        self,
        interface: AbsProtocol,
        device: AbsProtocol.Device,
        executor: ThreadPoolExecutor,
        loop: asyncio.AbstractEventLoop,
    ) -> None:
        """
        Constructor

        Args:
            serial: connected Serial object
            executor: a thread pool executor
            loop: event loop
        """
        self._interface = interface
        self._device = device
        self._executor = executor
        self._loop = loop
        self._supported_wavelengths: Optional[list[int]] = None
        self._device_handle: Optional[int] = None
        self._current_config: Optional[AbsProtocol.MeasurementConfig] = None

    def _cleanup(self) -> None:
        self._device_handle = None

    def _open(self) -> None:
        err, device_handle = self._interface.byonoy_open_device(self._device)
        if err.name != "BYONOY_ERROR_NO_ERROR":
            raise RuntimeError(f"Error opening device: {err}")
        self._device_handle = device_handle

    def _free(self) -> None:
        if self._device_handle:
            self._interface.byonoy_free_device(self._device_handle)
            self._cleanup()

    def verify_device_handle(self) -> int:
        assert self._device_handle is not None, RuntimeError(
            "Device handle not set up."
        )
        return self._device_handle

    def _raise_if_error(
        self,
        err_name: ErrorCodeNames,
        msg: str = "Error occurred: ",
    ) -> None:
        if err_name != "BYONOY_ERROR_NO_ERROR":
            raise RuntimeError(msg, err_name)

    def _get_device_information(self) -> AbsProtocol.DeviceInfo:
        handle = self.verify_device_handle()
        err, device_info = self._interface.byonoy_get_device_information(handle)
        self._raise_if_error(err.name, "Error getting device information: ")
        return device_info

    def _get_device_status(self) -> AbsProtocol.DeviceState:
        handle = self.verify_device_handle()
        err, status = self._interface.byonoy_get_device_status(handle)
        self._raise_if_error(err.name, "Error getting device status: ")
        return status

    def _get_slot_status(self) -> AbsProtocol.SlotState:
        handle = self.verify_device_handle()
        err, slot_status = self._interface.byonoy_get_device_slot_status(handle)
        self._raise_if_error(err.name, "Error getting slot status: ")
        return slot_status

    def _get_lid_status(self) -> bool:
        handle = self.verify_device_handle()
        lid_on: bool
        err, lid_on = self._interface.byonoy_get_device_parts_aligned(handle)
        self._raise_if_error(err.name, "Error getting lid status: ")
        return lid_on

    def _get_supported_wavelengths(self) -> List[int]:
        handle = self.verify_device_handle()
        wavelengths: List[int]
        err, wavelengths = self._interface.byonoy_abs96_get_available_wavelengths(
            handle
        )
        self._raise_if_error(err.name, "Error getting available wavelengths: ")
        self._supported_wavelengths = wavelengths
        return wavelengths

    def _initialize_measurement(self, conf: AbsProtocol.MeasurementConfig) -> None:
        handle = self.verify_device_handle()
        err = self._interface.byonoy_abs96_initialize_single_measurement(handle, conf)
        self._raise_if_error(err.name, "Error initializing measurement: ")
        self._current_config = conf

    def _single_measurement(self, conf: AbsProtocol.MeasurementConfig) -> List[float]:
        handle = self.verify_device_handle()
        measurements: List[float]
        err, measurements = self._interface.byonoy_abs96_single_measure(handle, conf)
        self._raise_if_error(err.name, "Error getting single measurement: ")
        return measurements

    def _set_sample_wavelength(self, wavelength: int) -> AbsProtocol.MeasurementConfig:
        if not self._supported_wavelengths:
            self._get_supported_wavelengths()
        assert self._supported_wavelengths
        if wavelength in self._supported_wavelengths:
            conf = self._interface.ByonoyAbs96SingleMeasurementConfig()
            conf.sample_wavelength = wavelength
            return conf
        else:
            raise ValueError(
                f"Unsupported wavelength: {wavelength}, expected: {self._supported_wavelengths}"
            )

    def _initialize(self, wavelength: int) -> None:
        conf = self._set_sample_wavelength(wavelength)
        self._initialize_measurement(conf)

    def _get_single_measurement(self, wavelength: int) -> List[float]:
        initialized = self._current_config
        assert initialized and initialized.sample_wavelength == wavelength
        return self._single_measurement(initialized)

    async def open(self) -> None:
        """
        Open the connection.

        Returns: None
        """
        return await self._loop.run_in_executor(
            executor=self._executor, func=self._open
        )

    async def close(self) -> None:
        """
        Close the connection

        Returns: None
        """
        await self._loop.run_in_executor(executor=self._executor, func=self._free)

    async def is_open(self) -> bool:
        """
        Check if connection is open.

        Returns: boolean
        """
        return self._device_handle is not None

    async def get_device_static_info(self) -> Dict[str, str]:
        return {
            "serial": self._device.sn,
            "model": "ABS96",
            "version": "1.0",
        }

    async def get_device_information(self) -> Dict[str, str]:
        device_info = await self._loop.run_in_executor(
            executor=self._executor, func=self._get_device_information
        )
        return {
            "serial_number": device_info.sn,
            "reference_number": device_info.ref_no,
            "version": device_info.version,
        }

    async def get_lid_status(self) -> AbsorbanceReaderLidStatus:
        lid_info = await self._loop.run_in_executor(
            executor=self._executor, func=self._get_lid_status
        )
        return (
            AbsorbanceReaderLidStatus.ON if lid_info else AbsorbanceReaderLidStatus.OFF
        )

    async def get_supported_wavelengths(self) -> list[int]:
        return await self._loop.run_in_executor(
            executor=self._executor, func=self._get_supported_wavelengths
        )

    async def initialize(self, wavelength: int) -> None:
        return await self._loop.run_in_executor(
            executor=self._executor, func=partial(self._initialize, wavelength)
        )

    async def get_single_measurement(self, wavelength: int) -> List[float]:
        return await self._loop.run_in_executor(
            executor=self._executor,
            func=partial(self._get_single_measurement, wavelength),
        )

    async def get_plate_presence(self) -> AbsorbanceReaderPlatePresence:
        return AbsorbanceReaderPlatePresence.UNKNOWN

    async def get_device_status(self) -> AbsorbanceReaderDeviceState:
        status = await self._loop.run_in_executor(
            executor=self._executor,
            func=self._get_device_status,
        )
        return self.convert_device_state(status.name)

    @staticmethod
    def convert_device_state(
        device_state: DeviceStateNames,
    ) -> AbsorbanceReaderDeviceState:
        state_map: Dict[DeviceStateNames, AbsorbanceReaderDeviceState] = {
            "BYONOY_DEVICE_STATE_UNKNOWN": AbsorbanceReaderDeviceState.UNKNOWN,
            "BYONOY_DEVICE_STATE_OK": AbsorbanceReaderDeviceState.OK,
            "BYONOY_DEVICE_STATE_BROKEN_FW": AbsorbanceReaderDeviceState.BROKEN_FW,
            "BYONOY_DEVICE_STATE_ERROR": AbsorbanceReaderDeviceState.ERROR,
        }
        return state_map[device_state]



"""The liquidProbe and tryLiquidProbe commands."""

from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Type, Union
from opentrons.protocol_engine.errors.exceptions import MustHomeError, TipNotEmptyError
from opentrons.types import MountType
from opentrons_shared_data.errors.exceptions import (
    PipetteLiquidNotFoundError,
)
from typing_extensions import Literal

from pydantic import Field

from ..types import DeckPoint
from .pipetting_common import (
    LiquidNotFoundError,
    LiquidNotFoundErrorInternalData,
    PipetteIdMixin,
    WellLocationMixin,
    DestinationPositionResult,
)
from .command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
    DefinedErrorData,
    SuccessData,
)

from ..errors.error_occurrence import ErrorOccurrence

if TYPE_CHECKING:
    from ..execution import MovementHandler, PipettingHandler
    from ..resources import ModelUtils


LiquidProbeCommandType = Literal["liquidProbe"]
TryLiquidProbeCommandType = Literal["tryLiquidProbe"]


# Both command variants should have identical parameters.
# But we need two separate parameter model classes because
# `command_unions.CREATE_TYPES_BY_PARAMS_TYPE` needs to be a 1:1 mapping.
class _CommonParams(PipetteIdMixin, WellLocationMixin):
    pass


class LiquidProbeParams(_CommonParams):
    """Parameters required for a `liquidProbe` command."""

    pass


class TryLiquidProbeParams(_CommonParams):
    """Parameters required for a `tryLiquidProbe` command."""

    pass


class LiquidProbeResult(DestinationPositionResult):
    """Result data from the execution of a `liquidProbe` command."""

    z_position: float = Field(
        ..., description="The Z coordinate, in mm, of the found liquid in deck space."
    )
    # New fields should use camelCase. z_position is snake_case for historical reasons.


class TryLiquidProbeResult(DestinationPositionResult):
    """Result data from the execution of a `tryLiquidProbe` command."""

    z_position: Optional[float] = Field(
        ...,
        description=(
            "The Z coordinate, in mm, of the found liquid in deck space."
            " If no liquid was found, `null` or omitted."
        ),
    )


_LiquidProbeExecuteReturn = Union[
    SuccessData[LiquidProbeResult, None],
    DefinedErrorData[LiquidNotFoundError, LiquidNotFoundErrorInternalData],
]
_TryLiquidProbeExecuteReturn = SuccessData[TryLiquidProbeResult, None]


class LiquidProbeImplementation(
    AbstractCommandImpl[LiquidProbeParams, _LiquidProbeExecuteReturn]
):
    """The implementation of a `liquidProbe` command."""

    def __init__(
        self,
        movement: MovementHandler,
        pipetting: PipettingHandler,
        model_utils: ModelUtils,
        **kwargs: object,
    ) -> None:
        self._movement = movement
        self._pipetting = pipetting
        self._model_utils = model_utils

    async def execute(self, params: _CommonParams) -> _LiquidProbeExecuteReturn:
        """Move to and liquid probe the requested well.

        Return the z-position of the found liquid.
        If no liquid is found, return a LiquidNotFoundError as a defined error.

        Raises:
            TipNotAttachedError: as an undefined error, if there is not tip attached to
                the pipette.
            TipNotEmptyError: as an undefined error, if the tip starts with liquid
                in it.
            MustHomeError: as an undefined error, if the plunger is not in a valid
                position.
        """
        pipette_id = params.pipetteId
        labware_id = params.labwareId
        well_name = params.wellName

        # _validate_tip_attached in pipetting.py is a private method so we're using
        # get_is_ready_to_aspirate as an indirect way to throw a TipNotAttachedError if appropriate
        self._pipetting.get_is_ready_to_aspirate(pipette_id=pipette_id)

        if self._pipetting.get_is_empty(pipette_id=pipette_id) is False:
            raise TipNotEmptyError(
                message="This operation requires a tip with no liquid in it."
            )

        if await self._movement.check_for_valid_position(mount=MountType.LEFT) is False:
            raise MustHomeError(
                message="Current position of pipette is invalid. Please home."
            )

        # liquid_probe process start position
        position = await self._movement.move_to_well(
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_name=well_name,
            well_location=params.wellLocation,
        )

        try:
            z_pos = await self._pipetting.liquid_probe_in_place(
                pipette_id=pipette_id,
                labware_id=labware_id,
                well_name=well_name,
                well_location=params.wellLocation,
            )
        except PipetteLiquidNotFoundError as e:
            return DefinedErrorData(
                public=LiquidNotFoundError(
                    id=self._model_utils.generate_id(),
                    createdAt=self._model_utils.get_timestamp(),
                    wrappedErrors=[
                        ErrorOccurrence.from_failed(
                            id=self._model_utils.generate_id(),
                            createdAt=self._model_utils.get_timestamp(),
                            error=e,
                        )
                    ],
                ),
                private=LiquidNotFoundErrorInternalData(
                    position=DeckPoint(x=position.x, y=position.y, z=position.z)
                ),
            )
        else:
            return SuccessData(
                public=LiquidProbeResult(
                    z_position=z_pos,
                    position=DeckPoint(x=position.x, y=position.y, z=position.z),
                ),
                private=None,
            )


class TryLiquidProbeImplementation(
    AbstractCommandImpl[TryLiquidProbeParams, _TryLiquidProbeExecuteReturn]
):
    """The implementation of a `tryLiquidProbe` command."""

    def __init__(
        self,
        movement: MovementHandler,
        pipetting: PipettingHandler,
        model_utils: ModelUtils,
        **kwargs: object,
    ) -> None:
        self._movement = movement
        self._pipetting = pipetting
        self._model_utils = model_utils

    async def execute(self, params: _CommonParams) -> _TryLiquidProbeExecuteReturn:
        """Execute a `tryLiquidProbe` command.

        `tryLiquidProbe` is identical to `liquidProbe`, except that if no liquid is
        found, `tryLiquidProbe` returns a success result with `z_position=null` instead
        of a defined error.
        """
        # We defer to the `liquidProbe` implementation. If it returns a defined
        # `liquidNotFound` error, we remap that to a success result.
        # Otherwise, we return the result or propagate the exception unchanged.

        original_impl = LiquidProbeImplementation(
            movement=self._movement,
            pipetting=self._pipetting,
            model_utils=self._model_utils,
        )
        original_result = await original_impl.execute(params)

        match original_result:
            case DefinedErrorData(
                public=LiquidNotFoundError(),
                private=LiquidNotFoundErrorInternalData() as original_private,
            ):
                return SuccessData(
                    public=TryLiquidProbeResult(
                        z_position=None,
                        position=original_private.position,
                    ),
                    private=None,
                )
            case SuccessData(
                public=LiquidProbeResult() as original_public, private=None
            ):
                return SuccessData(
                    public=TryLiquidProbeResult(
                        position=original_public.position,
                        z_position=original_public.z_position,
                    ),
                    private=None,
                )


class LiquidProbe(
    BaseCommand[LiquidProbeParams, LiquidProbeResult, LiquidNotFoundError]
):
    """The model for a full `liquidProbe` command."""

    commandType: LiquidProbeCommandType = "liquidProbe"
    params: LiquidProbeParams
    result: Optional[LiquidProbeResult]

    _ImplementationCls: Type[LiquidProbeImplementation] = LiquidProbeImplementation


class TryLiquidProbe(
    BaseCommand[TryLiquidProbeParams, TryLiquidProbeResult, ErrorOccurrence]
):
    """The model for a full `tryLiquidProbe` command."""

    commandType: TryLiquidProbeCommandType = "tryLiquidProbe"
    params: TryLiquidProbeParams
    result: Optional[TryLiquidProbeResult]

    _ImplementationCls: Type[
        TryLiquidProbeImplementation
    ] = TryLiquidProbeImplementation


class LiquidProbeCreate(BaseCommandCreate[LiquidProbeParams]):
    """The request model for a `liquidProbe` command."""

    commandType: LiquidProbeCommandType = "liquidProbe"
    params: LiquidProbeParams

    _CommandCls: Type[LiquidProbe] = LiquidProbe


class TryLiquidProbeCreate(BaseCommandCreate[TryLiquidProbeParams]):
    """The request model for a `tryLiquidProbe` command."""

    commandType: TryLiquidProbeCommandType = "tryLiquidProbe"
    params: TryLiquidProbeParams

    _CommandCls: Type[TryLiquidProbe] = TryLiquidProbe



"""Command models to initialize an Absorbance Reader."""
from __future__ import annotations
from typing import Optional, Literal, TYPE_CHECKING
from typing_extensions import Type

from pydantic import BaseModel, Field

from ..command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ...errors.error_occurrence import ErrorOccurrence

if TYPE_CHECKING:
    from opentrons.protocol_engine.state import StateView
    from opentrons.protocol_engine.execution import EquipmentHandler


InitializeCommandType = Literal["absorbanceReader/initialize"]


class InitializeParams(BaseModel):
    """Input parameters to initialize an absorbance reading."""

    moduleId: str = Field(..., description="Unique ID of the absorbance reader.")
    sampleWavelength: int = Field(..., description="Sample wavelength in nm.")


class InitializeResult(BaseModel):
    """Result data from initializing an aborbance reading."""


class InitializeImpl(
    AbstractCommandImpl[InitializeParams, SuccessData[InitializeResult, None]]
):
    """Execution implementation of initializing an Absorbance Reader."""

    def __init__(
        self,
        state_view: StateView,
        equipment: EquipmentHandler,
        **unused_dependencies: object,
    ) -> None:
        self._state_view = state_view
        self._equipment = equipment

    async def execute(
        self, params: InitializeParams
    ) -> SuccessData[InitializeResult, None]:
        """Initiate a single absorbance measurement."""
        abs_reader_substate = self._state_view.modules.get_absorbance_reader_substate(
            module_id=params.moduleId
        )
        # Allow propagation of ModuleNotAttachedError.
        abs_reader = self._equipment.get_module_hardware_api(
            abs_reader_substate.module_id
        )

        if abs_reader is not None:
            await abs_reader.set_sample_wavelength(wavelength=params.sampleWavelength)

        return SuccessData(
            public=InitializeResult(),
            private=None,
        )


class Initialize(BaseCommand[InitializeParams, InitializeResult, ErrorOccurrence]):
    """A command to initialize an Absorbance Reader."""

    commandType: InitializeCommandType = "absorbanceReader/initialize"
    params: InitializeParams
    result: Optional[InitializeResult]

    _ImplementationCls: Type[InitializeImpl] = InitializeImpl


class InitializeCreate(BaseCommandCreate[InitializeParams]):
    """A request to execute an Absorbance Reader measurement."""

    commandType: InitializeCommandType = "absorbanceReader/initialize"
    params: InitializeParams

    _CommandCls: Type[Initialize] = Initialize



from typing import (
    Dict,
    Protocol,
    List,
    Literal,
    Tuple,
    runtime_checkable,
    TypeVar,
)

Response = TypeVar("Response")

ErrorCodeNames = Literal[
    "BYONOY_ERROR_NO_ERROR",
    "BYONOY_ERROR_UNKNOWN_ERROR",
    "BYONOY_ERROR_DEVICE_CLOSED",
    "BYONOY_ERROR_INVALID_ARGUMENT",
    "BYONOY_ERROR_NO_MEMORY",
    "BYONOY_ERROR_UNSUPPORTED_OPERATION",
    "BYONOY_ERROR_DEVICE_COMMUNICATION_FAILURE",
    "BYONOY_ERROR_DEVICE_OPERATION_FAILED",
    "BYONOY_ERROR_DEVICE_OPEN_PREFIX",
    "BYONOY_ERROR_DEVICE_NOT_FOUND",
    "BYONOY_ERROR_DEVICE_TOO_NEW",
    "BYONOY_ERROR_DEVICE_ALREADY_OPEN",
    "BYONOY_ERROR_FIRMWARE_UPDATE_ERROR_PREFIX",
    "BYONOY_ERROR_FIRMWARE_UPDATE_FILE_NOT_FOUND",
    "BYONOY_ERROR_FIRMWARE_UPDATE_FILE_NOT_VALID",
    "BYONOY_ERROR_FIRMWARE_UPDATE_FAILED",
    "BYONOY_ERROR_FILE_ERROR_PREFIX",
    "BYONOY_ERROR_FILE_WRITE_ERROR",
    "BYONOY_ERROR_MEASUTEMNT_ERROR_PREFIX",
    "BYONOY_ERROR_MEASUTEMNT_SLOT_NOT_EMPTY",
    "BYONOY_ERROR_NOT_INITIALIZED",
    "BYONOY_ERROR_INTERNAL",
]

SlotStateNames = Literal[
    "BYONOY_SLOT_UNKNOWN",
    "BYONOY_SLOT_EMPTY",
    "BYONOY_SLOT_OCCUPIED",
    "BYONOY_SLOT_UNDETERMINED",
]

DeviceStateNames = Literal[
    "BYONOY_DEVICE_STATE_UNKNOWN",
    "BYONOY_DEVICE_STATE_OK",
    "BYONOY_DEVICE_STATE_BROKEN_FW",
    "BYONOY_DEVICE_STATE_ERROR",
]


@runtime_checkable
class AbsorbanceHidInterface(Protocol):
    @runtime_checkable
    class Device(Protocol):
        sn: str

    @runtime_checkable
    class ErrorCode(Protocol):
        __members__: Dict[ErrorCodeNames, int]
        name: ErrorCodeNames
        value: int

    @runtime_checkable
    class SlotState(Protocol):
        __members__: Dict[SlotStateNames, int]
        name: SlotStateNames
        value: int

    @runtime_checkable
    class MeasurementConfig(Protocol):
        sample_wavelength: int

    @runtime_checkable
    class DeviceInfo(Protocol):
        ref_no: str
        sn: str
        version: str

    @runtime_checkable
    class DeviceState(Protocol):
        __members__: Dict[DeviceStateNames, int]
        name: DeviceStateNames
        value: int

    def ByonoyAbs96SingleMeasurementConfig(self) -> MeasurementConfig:
        ...

    def byonoy_open_device(self, device: Device) -> Tuple[ErrorCode, int]:
        ...

    def byonoy_free_device(self, device_handle: int) -> Tuple[ErrorCode, bool]:
        ...

    def byonoy_get_device_information(
        self, device_handle: int
    ) -> Tuple[ErrorCode, DeviceInfo]:
        ...

    def byonoy_get_device_status(
        self, device_handle: int
    ) -> Tuple[ErrorCode, DeviceState]:
        ...

    def byonoy_get_device_slot_status(
        self, device_handle: int
    ) -> Tuple[ErrorCode, SlotState]:
        ...

    def byonoy_get_device_parts_aligned(
        self, device_handle: int
    ) -> Tuple[ErrorCode, bool]:
        ...

    def byonoy_abs96_get_available_wavelengths(
        self, device_handle: int
    ) -> Tuple[ErrorCode, List[int]]:
        ...

    def byonoy_abs96_initialize_single_measurement(
        self, device_handle: int, conf: MeasurementConfig
    ) -> ErrorCode:
        ...

    def byonoy_abs96_single_measure(
        self, device_handle: int, conf: MeasurementConfig
    ) -> Tuple[ErrorCode, List[float]]:
        ...

    def byonoy_available_devices(self) -> List[Device]:
        ...
